/*
    $Id: vpd.c,v 1.3 1999/07/11 01:51:06 james Exp $

    vpd, Granite Communications BS5 Base Station Communication Daemon
    Copyright (C) 1999  James Cameron  (quozl@us.netrek.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef lint
static char vcid[] = "$Id: vpd.c,v 1.3 1999/07/11 01:51:06 james Exp $";
#endif /* lint */

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "vpd.h"
#include "bsf.h"
#include "bsc.h"
#include "bst.h"

#define DEVICE "/dev/bs5"
#define MINPDU 2
#define MAXPDU 2083

static int fd;			/* file descriptor of base station	*/
static int flag_dump  = 0;	/* if we should dump packets on screen	*/
static int flag_trace = 0;	/* if we should display low level data	*/

static char *id ( const struct id *it )
{
  static char buffer[13];
  sprintf(buffer, "%02.2x%02.2x%02.2x%02.2x%02.2x%02.2x",
	  it->id[0], it->id[1], it->id[2], it->id[3], it->id[4], it->id[5]);
  return buffer;
}

static char *decode ( unsigned char byte )
{
  static char out[6];
  
  switch (byte) {
  case STX  : return "STX ";
  case ETX  : return "ETX ";
  case ACK0 : return "ACK0 ";
  case ACK1 : return "ACK1 ";
  case DLE  : return "DLE ";
  case NAK0 : return "NAK0 ";
  case NAK1 : return "NAK1 ";
  default:  sprintf(out, "%02.2x ", byte); return out;
  }
}

static void dump ( char *format, unsigned char *start, int length )
{
  char buffer[MAXPACKET];
  unsigned char *in = start;
  char *out = buffer;
  unsigned char *end = start + length;
  
  for(;in<end;in++) {
    strcpy(out, decode(*in));
    out += strlen(out);
  }
  *out++ = '\0';
  printf(format, buffer);
}
    
void ack ( unsigned char ctl )
{
  int status;
  unsigned char acknowledgement[2];
  
  acknowledgement[0] = DLE;
  if (ctl && 0x80)
    acknowledgement[1] = ACK1;
  else
    acknowledgement[1] = ACK0;
  
  if (flag_dump) printf(">bs: ACK\n");
  if (flag_trace) dump(">port: %s\n", acknowledgement, 2);
  
  status = write(fd, acknowledgement, 2);
  if (status < 0) {
    perror("write");
  }
}

nak ( unsigned char ctl )
{
  int status;
  unsigned char acknowledgement[2];
  
  acknowledgement[0] = DLE;
  if ( ctl && 0x80 )
    acknowledgement[1] = NAK1;
  else
    acknowledgement[1] = NAK0;
  
  if (flag_dump) printf(">bs: NAK\n");
  if (flag_trace) dump(">port: %s\n", acknowledgement, 2);
        
  status = write(fd, acknowledgement, 2);
  if (status < 0) {
    perror("write");
  }
}

#define RX_TYPE_DATA 0
#define RX_TYPE_NAK 1
#define RX_TYPE_ACK 2
#define RX_TYPES 3
int counts[RX_TYPES];

struct rx {
  unsigned char type;
  unsigned int length;
  unsigned char data[MAXPDU];
  unsigned char null;
};

#define PARSE_STATE_NORMAL 0

/*
 * Parse an incoming byte from the base station.  Return NULL if a complete 
 * packet has not yet been received, otherwise return the address of a 
 * returned packet.
 */
struct rx *parse ( unsigned char byte )
{
  static unsigned char dle = 0;	/* context, prior byte was DLE	*/
  static unsigned char stx = 0;	/* context, prior byte was STX	*/
  static unsigned char etx = 0;	/* context, prior byte was ETX	*/
  static unsigned char sum = 0;	/* context, prior byte was sum	*/
  
  static struct rx rx;		/* return structure		*/
  static unsigned char ctl;	/* control byte after STX	*/
  static unsigned char cks[2];	/* remote's checksum bytes	*/
  static int acc;		/* our checksum accumulator	*/
  static int dpr;		/* offset to data in rx.data	*/

  /*
   * The Granite Communications documentation I have for the Base Station
   * protocol is less than clear on how DLE is used.  The obvious
   * interpretation was that a DLE would "escape" anything within a data
   * packet that was not protocol related.  However the engineers have
   * implemented it so that a DLE _must_ prefix any protocol byte, and so
   * the data portion of the packet is pretty much raw. - quozl
   */
    
  /* did we just previously receive a DLE? */
  if (dle) {
    dle = 0;
    switch (byte) {
    case DLE:		/* was it a DLE after DLE?	*/
      break;		/* treat as data		*/
    case STX:
      stx = 1;
      return NULL;
    case ETX:
      etx = 1;
      return NULL;
    case ACK0:
    case ACK1:
      rx.type = RX_TYPE_ACK;
      return &rx;
    case NAK0:
    case NAK1:
      rx.type = RX_TYPE_NAK;
      return &rx;
    default:
      printf("parse: byte %02.2x after DLE dropped\n", byte);
      return NULL;
    }
  } else {
    /* if we didn't previously see a DLE, and we do now ... */
    if (byte == DLE) {
      dle = 1;		/* set flag to say seen one	*/
      return NULL;
    }
  }
  
  /* did we just receive STX?  if so, new byte is CTL */
  if (stx) {
    ctl = byte;		/* save the control byte for ACK	*/
    acc = ctl;		/* start the accumulated sum for check	*/
    stx = 0;		/* reset the stx flag			*/
    dpr = 0;		/* reset the data pointer		*/
    return NULL;
  }
  
  /* did we just receive ETX?  if so, new byte is SUM */
  if (etx) {
    acc += ETX;		/* include the ETX in the accumulator	*/
    cks[0] = byte;	/* save the first byte of the checksum	*/
    etx = 0;		/* turn off flag			*/
    sum = 1;		/* expect second byte of checksum	*/
    return NULL;
  }
    
  /* are we expecting second byte of checksum? */
  if (sum) {
    int acb[2];
    
    cks[1] = byte;	/* save second byte of checksum		*/
    sum = 0;		/* no longer expecting data		*/
    
    /* split out our calculated checksum */
    acb[0] = acc & 0xff;
    acb[1] = ( acc & 0xff00 ) >> 8;
    
    /* check checksum */
    if ( ( cks[0] == acb[0] ) && ( cks[1] == acb[1] ) ) {
      /* generate ACK */
      ack ( ctl );
      rx.length = dpr;		/* set length of pdu		*/
      rx.type = RX_TYPE_DATA;	/* set type of return struct	*/
      return &rx;		/* return the data packet	*/
    } else {
      /* generate NAK */
      printf("vpd: we nak, got %02.2x %02.2x expected %02.2x %02.2x\n",
	     cks[0], cks[1], acb[0], acb[1]);
      nak(ctl);
      return NULL;
    }
  }
  
  /* discard data after maximum PDU length */
  if ( dpr > MAXPDU ) return NULL;
  
  /* otherwise it's pdu data */
  acc += byte;			/* accumulate in checksum	*/
  rx.data[dpr++] = byte;	/* save in return structure	*/
  
  return NULL;
}

/*
 * Check the file descriptor attached to the serial port for data, and return
 * a zero after one second if no data has arrived.  If data does arrive,
 * return the number of bytes written into the buffer provided.
 */
int qio ( unsigned char *packet )
{
  int status;
  struct timeval timeout;
  fd_set readfds;

  timeout.tv_sec = 1;	/* wait for one second for data	*/
  timeout.tv_usec = 0;

  if (flag_trace > 1) printf ( "qio: start select\n" );

  FD_ZERO(&readfds);	/* clear the file descriptor mask	*/
  FD_SET(fd,&readfds);	/* set the bit for the serial port fd	*/
  status = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout);

  if (status == 0) {
    if (flag_trace > 1) printf ( "qio: select returned zero\n" );
    return 0;
  }

  if (status < 0) {
    if (flag_trace > 1) printf ( "qio: select returned error\n" );
    perror("select");
    return 0;
  }

  if (flag_trace > 1) printf ( "qio: select returned data\n" );

  /* check that select() set the bit for our file descriptor */
  if (FD_ISSET(fd,&readfds)) {
    if (flag_trace > 1) printf ( "qio: start read\n" );
    
    /* read the data into the packet buffer */
    status = read(fd, packet, MAXPACKET);
    if (flag_trace > 1) printf("qio: read returns length %d\n", status);
    if (status < 0) {
      perror("read");
      return 0;
    }
  }

  /* return the number of bytes that were read */
  return status;
}

/*
 * Purge the type-ahead buffer (input queue) of the serial port.
 */
void purge ( )
{
  int status;
  
  status = tcflush(fd, TCIFLUSH);
  if (status < 0) {
    perror("tcflush");
  }
}

static int send ( unsigned char *buffer, int length )
{
  int status;
  unsigned char packet[MAXPACKET];
  unsigned char *i, *o;
  int sum;
  int x;
  static unsigned char ctl = 0;
  
  if (flag_dump) dump(">bs: %s\n", buffer, length);
  
  i = buffer;
  o = packet;
  
  sum = 0;
  *o++ = DLE;
  *o++ = STX;
  *o = ctl; sum += *o++;
  ctl = ( ctl == 0 ) ? 0x80 : 0;
  
  for(x=0;x<length;x++) { *o = *i++; sum += *o++; }
  
  *o++ = DLE;
  *o = ETX; sum += *o++;
  *o++ = sum & 0xff;
  *o++ = ( sum & 0xff00 ) >> 8;
  
  if (flag_trace) dump(">port: %s\n", packet, o - packet);
  
  status = write(fd, packet, o - packet);
  if (status < 0) {
    perror("write");
  }
}

void berror ( struct bs_s *reply )
{
  char *text;
  
  switch ( reply->cmderr ) {
  case 0x00: 
    if (flag_dump == 0) return;
    text = "OK";                          
    break;
  case 0x24: text = "base not initialised";        break;
  case 0x28: text = "base is initialised";         break;
  case 0x27: text = "invalid logical unit number"; break;
  case 0x2a: text = "message queue full";          break;
  case 0x25: text = "no base memory";              break;
  case 0x21: text = "RF not disabled";             break;
  case 0x22: text = "RF not enabled";              break;
  case 0x23: text = "syntax error";                break;
  case 0x26: text = "unknown error";               break;
  case 0x20: text = "unrecognised command";        break;
  case 0x29: text = "RF enabled not allowed";      break;
  default:
    {
      static char reason[80];
      sprintf ( reason, "unknown error 0x%02.2x", reply->cmderr );
    }
  }
  printf ( "Status: %s\n", text );
}

void send_block (unsigned char rlun, 
		 const unsigned char *buffer, 
		 unsigned short int length)
{
  unsigned int size;
  unsigned char ours[MAXPDU+sizeof(struct bs_send_block)];
  struct bs_send_block *block;
  
  size = sizeof(struct bs_send_block) + length - 1;
  block = (struct bs_send_block *) ours;
  block->type = bs_send_block.type;
  block->tag  = BS_SEND_BLOCK_TAG_NORMAL;
  block->rlun = rlun;
  block->len  = length;
  memcpy(&block->pdu, buffer, length);
  
  send(ours, size);
  return;
}

unsigned char onelun = 0;
unsigned char twolun = 0;

void echo ( const struct bs_rece_block *block )
{
    unsigned char it = block->rlun;
    unsigned char ot = 0;
    
    if ( it == onelun ) ot = twolun;
    if ( it == twolun ) ot = onelun;
    
    if ( ot == 0 ) 
    {
        printf ( "accept: ignored data from lun %02.2x\n", block->rlun );
        return;
    }
    
    /* send block to other terminal */
    send_block ( ot, ( unsigned char * ) &block->pdu, block->len );
}

static void accept ( const struct bs_rece_block *block )
{
  char *response;
  response = bsf(block->rlun, &block->pdu, block->len);
  if (response != NULL) {
    send_block(block->rlun, (unsigned char *) response, strlen(response));
    return;
  }
  response = bsc(block->rlun, &block->pdu, block->len);
  if (response != NULL) {
    send_block(block->rlun, (unsigned char *) response, strlen(response));
    return;
  }
  bst(block->rlun, &block->pdu, block->len);
}

void logon ( const struct bs_logon *logon )
{
  char *it = id(&logon->padd);
  char *one = "f0db16000000";
  char *two = "4ece16000000";
  
  printf ( "logon: heard from %s, lun %02.2x\n", it, logon->rlun );
  bsf_logon ( it, logon->rlun );
  if ( ! strcmp ( it, one ) ) { onelun = logon->rlun; return; }
  if ( ! strcmp ( it, two ) ) { twolun = logon->rlun; return; }
  printf ( "logon: denied from %s\n", id );
}

void logoff ( struct bs_logoff *logoff )
{
  bsf_logoff ( logoff->rlun );
  printf ( "logoff: goodbye to lun %02.2x\n", logoff->rlun );
  if ( onelun == logoff->rlun ) { onelun = 0; return; }
  if ( twolun == logoff->rlun ) { twolun = 0; return; }
  printf ( "logoff: ignored from lun %02.2x\n", logoff->rlun );
}

handle ( const unsigned char *buffer, int length )
{
  switch (buffer[0]) {
  case BS_RECE_BLOCK:
    accept((struct bs_rece_block *) buffer);
    break;
  case BS_LOGON:
    logon((struct bs_logon *) buffer);
    break;
  case BS_LOGOFF:
    logoff((struct bs_logoff * ) buffer);
    break;
  case BS_STATUS:
    berror((struct bs_s *) buffer);
    break;
  default:
    printf("handle: unrecognised packet from base\n");
    break;
  }
}

#define BS_NIGHTLY_HOUR 2	/* 2AM in the morning */

static bs_nightly()
{
    static int ticks = 0;
    static time_t was = 0;
    time_t now;
    struct tm *tm;
    
    /* check for time every just-under-a-minute */
    if ( ticks++ < 55 ) return;
    
    /* following code therefore executes every 55 seconds or so */
    ticks = 0;
    
    /* if it's not yet time, return */
    now = time ( NULL );
    tm = localtime ( &now );
    if ( tm->tm_hour != BS_NIGHTLY_HOUR ) return;
    
    /* ignore if we recently (last two hours) did what we want to do */
    if ( ( now - was ) < ( 2 * 60 * 60 ) ) return;
    
    was = now;
    
    /* do nightly activities */
    
#ifdef VMS
    /*
    **  Do Butler log dump to disk.  See dv:[butler]bu.c for more details.
    **  Give log dumper a time in the previous day for stamping the file with,
    **  less ten minutes, just in case.  It's the date the dumper uses.
    */
    bu_open();
    bu_log_dump ( now - ( BS_NIGHTLY_HOUR * 60 * 60 + 600 ) );
    bu_close();
#endif
    
    return;
}

void idle ( )
{
  int bytes, i;
  static unsigned char packet[MAXPACKET];	/* our buffer space	*/
  struct rx *rx;
  
  bytes = qio(packet);
  if (bytes == 0) {
    bs_nightly();
    return;
  }
  if (flag_trace) dump("port>: %s\n", packet, bytes);
    
  for(i=0;i<bytes;i++) {
    rx = parse(packet[i]);
    if (rx == NULL) continue;
    switch (rx->type) {
    case RX_TYPE_DATA:
      counts[rx->type]++;
      if (flag_dump) dump("bs>: %s\n", rx->data, rx->length);
      handle(rx->data, rx->length);
      continue;
    case RX_TYPE_ACK:
      counts[rx->type]++;
      if (flag_dump) printf("bs>: ACK\n");
      continue;
    case RX_TYPE_NAK:
      counts[rx->type]++;
      if (flag_dump) printf("bs>: NAK\n");
      continue;
    }
  }
  return;
}

void await ( int nd, int na )
{
  int i, od, oa;
    
  if (flag_trace > 1) printf("await: want %d data and %d acks\n", nd, na);
  if (nd == 0&&na == 0) return;
  
  od = counts[RX_TYPE_DATA];
  oa = counts[RX_TYPE_ACK];
  
  while(1) {
    idle();
    if (flag_trace > 1)
      printf("await: have %d data and %d acks\n",
	     counts[RX_TYPE_DATA] - od,
	     counts[RX_TYPE_ACK]  - oa);
    if (nd != 0) if ((counts[RX_TYPE_DATA] - od) < nd) continue;
    if (na != 0) if ((counts[RX_TYPE_ACK]  - oa) < na) continue;
    return;
  }
}

void reset ()
{
  if (flag_trace) printf("RESET begins\n");
  send((unsigned char *)&bs_reset, sizeof(struct bs_reset));
  await(0,1);
  if (flag_trace) printf("RESET done\n");
}

void initialise_base ()
{
  if (flag_trace) printf("INITIALISE BASE begins\n");
  send((unsigned char *)&bs_initialise_base, 
       sizeof(struct bs_initialise_base));
  await(1,1);
  if (flag_trace) printf("INITIALISE BASE done\n");
}

void enable_rf ()
{
  if (flag_trace) printf("ENABLE RF begins\n");
  send((unsigned char *)&bs_enable_rf, sizeof(struct bs_enable_rf));
  await(1,1);
  if (flag_trace) printf("ENABLE RF done\n");
}

void disable_rf ()
{
  if (flag_trace) printf("DISABLE RF begins\n");
  send((unsigned char *)&bs_disable_rf, sizeof(struct bs_disable_rf));
  await(1,1);
  if (flag_trace) printf("DISABLE RF ends\n");
}

void terminate ( )
{
  purge();
  disable_rf();
}

void usage()
{
  fprintf(stderr, 
	  "\n%s\n"
          "vpd, Copyright (C) 1999 James Cameron\n"
          "vpd comes with ABSOLUTELY NO WARRANTY; for details see source.\n"
          "This is free software, and you are welcome to redistribute it\n"
          "under certain conditions; see source for details.\n\n", vcid );
  fprintf(stderr,
	  "Usage: vpd [-h] [-f device] [-d] [-t] [-r]\n"
	  "       -h  display usage and exit\n"
	  "       -f  specify the serial port to use, default " DEVICE "\n"
	  "       -d  increment data dumping verbosity\n"
	  "       -t  increment program trace verbosity\n"
	  "       -r  stop a running vpd cleanly and then exit\n" );
}

int main ( int argc, char **argv )
{
  struct termios termios;
  int i;
  char *device = DEVICE;
  int quick = 0;
  
  for(i=1;i<argc;i++) {
    if (!strcmp(argv[i], "-h")) {
      usage();
      exit(1);
    }
    if (!strcmp(argv[i], "-f")) {
      if (i++ == argc) break;
      device = argv[i];
      continue;
    }
    if (!strcmp(argv[i], "-d")) { flag_dump++;  continue; }
    if (!strcmp(argv[i], "-t")) { flag_trace++; continue; }
    if (!strcmp(argv[i], "-q")) { quick++;    continue; }
    if (!strcmp(argv[i], "-r")) {
      FILE *pid = fopen("vpd.pid","r");
      pid_t value;
      if (pid == NULL) {
	perror("fopen: vpd.pid");
	exit(2);
      }
      fscanf(pid,"%d",&value);
      fclose(pid);
      exit(kill(value, SIGINT));
    }
    usage();
    fprintf(stderr, "\nUnrecognised option %s\n", argv[i]);
    exit(1);
  }
  
  /* write our pid file */
  {
    FILE *pid;
    if (flag_trace) printf("vpd: write pid file\n");
    pid = fopen("vpd.pid","w");
    fprintf(pid,"%d",getpid());
    fclose(pid);
  }
  
  bsf_load();
  
  /*
   * Open the serial port attached to the base station.
   *
   * O_RDWR	we want to read and write to it,
   * O_NOCTTY	we do not want it to be our controlling tty,
   * O_NONBLOCK	don't block waiting for data, return immediately on read(),
   * O_SYNC	perform write() synchronously, return on last byte sent.
   */
  fd = open(device, O_RDWR|O_NOCTTY|O_NONBLOCK|O_SYNC);
  if (fd < 0) {
    perror("open");
    exit(1);
  }

  if (tcgetattr(fd, &termios) < 0) {
    perror ( "tcgetattr" );
    exit(1);
  }

  cfsetospeed(&termios, B19200);
  cfsetispeed(&termios, B19200);
  cfmakeraw(&termios);

  if (tcsetattr(fd, TCSANOW, &termios) < 0) {
    perror ( "tcsetattr" );
    exit(1);
  }

  for(i=0;i<RX_TYPES;i++) counts[i] = 0;
  
  /*
   * Initialisation of the base station is described in the documentation
   * as requiring _two_ reset commands separated by five seconds, presumably
   * in case the first reset command arrived when the base was expecting a
   * packet with a different CTL field.  One second seems to work fine.
   *
   * [Ref: Granite Communications, Base to Host Communications Protocol, 
   * Reference Guide, Rev B 10/01/93]
   *
   * However, immediately after power up, the two reset commands and the
   * two one second waits are not required.  For the moment, "-q -q" may
   * be specified on the command line to skip this.  Note that the base
   * refuses to initialise or enable RF unless it has been first reset.
   */
  purge();		/* flush serial port input queue		*/

  if(quick < 1) {
    reset();		/* request base station reset			*/
    sleep(1);		/* sleep for one second				*/
  }

  if(quick < 2) {
    reset();		/* request base station reset			*/
    sleep(1);		/* sleep for one second				*/
  }

  initialise_base();	/* initialise the base station			*/
  signal(SIGINT,exit);	/* request a SIGINT call exit()			*/
  atexit(terminate);	/* on exit call our termination function	*/
  enable_rf();		/* enable radio frequency subsystem		*/
  while(1) idle(0);	/* process data until SIGINT			*/
}
